import time
import struct

from tqdm import trange
from udp_eth import UDP_, Ethernet
from printer import pass_print, error_print

class upgrader:
    def __init__(self):
        self.udp = UDP_()
        self.eth = Ethernet()
        self.fw_size = None

    def fw_content_generator(self, fw_path=None):
        if fw_path is None:
            fw_path = input("IMU330ra firmware path: \n")
            fw_path = fw_path.replace('"', '')
        fw_file = open(fw_path, "rb")
        fw_data = fw_file.read()
        self.fw_size = len(fw_data)
        return fw_data

    def extra_j2b(self, timeout=20):
        src_mac = '38:14:28:3f:12:e2'
        dst_mac = 'FF:FF:FF:FF:FF:FF'
        iface = '以太网'

        self.eth.src_mac = src_mac
        self.eth.dst_mac = dst_mac
        self.eth.iface = iface

        JI_MAC = [0x01, 0xaa]
        ji_mac_cmd = self.eth.build_packet(JI_MAC ,[])
        input('Please repower device, trying to let device jump to bootloader...')
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                resp = self.eth.write_read_response(JI_MAC, ji_mac_cmd)
                if resp is not None:
                    return True
                else:
                    time.sleep(0.1)
            else:
                error_print('Device did not in app mode')
                return False

    def extra_ping(self, timeout=20):
        src_mac = '38:14:28:3f:12:e2'
        dst_mac = 'FF:FF:FF:FF:FF:FF'
        iface = '以太网'

        self.eth.src_mac = src_mac
        self.eth.dst_mac = dst_mac
        self.eth.iface = iface

        PK_MAC = [0x01, 0xcc]
        ping_mac_cmd = self.eth.build_packet(PK_MAC ,[])
        input('trying to ping device')
        start_time = time.time()
        while True:
            if time.time() - start_time < timeout:
                resp = self.eth.write_read_response(PK_MAC, ping_mac_cmd, 0.1)
                if resp is not None:
                    pass_print('reconnect success')
                    return True
                else:
                    time.sleep(0.1)
            else:
                return False
        
    def ping_device(self):
        PK = [0x50, 0x4B]
        cnt = 1
        payload = b'\x00\x00'
        pk_cmd = self.udp.build_packet(PK, cnt, payload)
        resp = self.udp.write_read_response(PK, pk_cmd, 10)
        if resp is not None:
            pass_print("Ping device Success")
            return True
        else:
            error_print("Ping device Failed")
            return False

    def jump2boot(self):
        JI = [0x4A, 0x49]
        cnt = 1
        payload = []
        jump2boot_cmd = self.udp.build_packet(JI, cnt, payload)
        while True:
            resp = self.udp.write_read_response(JI, jump2boot_cmd, 0.1)
            if resp is not None:
                break
        if resp is not None:
            pass_print("Jump to boot Success")
            return True
        else:
            error_print("Jump to boot Failed")
            return False

    def jump2app(self):
        JB = [0x02, 0xaa]
        payload = []
        jump2app_cmd = self.eth.build_packet(JB, payload)
        resp = self.eth.write_read_response(JB, jump2app_cmd, 1)
        if resp is not None:
            pass_print("Jump to app Success")
            return True
        else:
            error_print("Jump to app Failed")
            return False

    def core_select(self, part_size, retry=1):
        CS = [0x04, 0xaa]
        update_core = bytes([0x43, 0x31])
        part_size = struct.pack('>I', part_size)
        app_info = bytes([0x69, 0x6e, 0x73])
        # payload = b'\x43\x31\x00\x02\x6f\x14\x69\x6e\x73'
        payload = update_core + part_size + app_info
        cs_cmd = self.eth.build_packet(CS, payload)
        for i in range(retry):
            resp = self.eth.write_read_response(CS, cs_cmd, 1)
            if resp is not None:
                break
        if resp is not None:
            # if resp[2] == b'C1S':
            #     pass_print("Core select Success")
            #     return True
            # else:
            #     error_print("Core select Failed")
            #     return False
            pass_print("Core select Success")
            return True
        else:
            error_print("Core select Failed")
            return False

    def write_app(self, fw_data):
        WA = [0x03, 0xaa]
        step = 192
        turn = int(len(fw_data)/step)
        for i in trange(turn):
            start_addr_payload = struct.pack('>I', i*step)
            fw_payload = fw_data[i*step:(i+1)*step]
            len_payload = struct.pack('>I', len(fw_payload))
            payload = start_addr_payload + len_payload + fw_payload
            wa_cmd = self.eth.build_packet(WA, payload)
            resp = self.eth.write_read_response(WA, wa_cmd, 10)
            if resp is not None:
                time.sleep(0.01)
            else:
                error_print("Write app failed")
                return False
        last_start_addr_payload = struct.pack('>I', turn*step)
        last_fw_payload = fw_data[turn*step:]
        for i in range(step - len(last_fw_payload)):
            last_fw_payload += b'\xff'
        last_len_payload = struct.pack('>I', len(last_fw_payload))
        last_payload = last_start_addr_payload + last_len_payload + last_fw_payload
        last_wa_cmd = self.eth.build_packet(WA, last_payload)
        last_resp = self.eth.write_read_response(WA, last_wa_cmd, 10)
        if last_resp is not None:
            time.sleep(0.01)
            print('Write app finished')
        else:
            error_print("Write app failed")
            return False
        return True

    def start(self):
        fw_path = input('Please input the firmware path:\n')
        fw_path = fw_path.replace('"', '')
        fw_data = self.fw_content_generator(fw_path) #cek-0: get FW size and data contents
        part_size = len(fw_data)
        ping_result = self.udp.find_device()
        if ping_result == False:
            self.eth.find_device()
        else:
            self.eth.eth_setting_update(self.udp.local_mac, self.udp.dev_mac, self.udp.iface)
            jump2boot_result = self.jump2boot()
            if jump2boot_result == False:
                return
            self.eth.__init__()
            self.eth.find_device()
        try:
            core_select_result = self.core_select(part_size, 10)
        except:
            self.eth.find_device()
            core_select_result = self.core_select(part_size, 10)
        if core_select_result == False:
            return
        if self.write_app(fw_data) == False:
            return
        jump2app_result = self.jump2app() # WARNING: If an error occurs during the previous upgrade, do not perform this operation. If you perform this operation forcibly it will brick your device
        if jump2app_result == False:
            return
    
    def rs(self):
        RS = [0x53, 0x52]
        cnt = 1
        payload = b'\x00\x00'
        self.udp.find_device()
        rs_cmd = self.udp.build_packet(RS, cnt, payload)
        resp = self.udp.write_read_response(RS, rs_cmd, 10)
        time.sleep(0.1)
    
if __name__ == '__main__': 
    u = upgrader()
    for i in range(1):
        u.__init__()
        # print(f'Turn{i+1}:')
        u.start()
        time.sleep(5)
        upgrade_result = u.udp.conmunication_check()
        if upgrade_result == False:
            error_print("Upgrade failed")
            break
        else:
            pass_print("Upgrade Success\n") 
